@@ -11,12 +11,8 @@
 //! the build script for subplot (by watching the share/ tree) and also a file
 //! which is included directly as the `EMBEDDED_FILES` constant in the
 //! `src/resources.rs` file.
-//!
-//! Finally the script checks for the `DEB_BUILD_OPTS` environment variable and
-//! if set, it causes a resource fallback path of `/usr/share/subplot` to be
-//! added to the resources search path automatically.
 
-use anyhow::Result;
+use anyhow::{anyhow, Result};
 use std::io::Write;
 use std::path::{Component, Path, PathBuf};
 use walkdir::WalkDir;
@@ -91,6 +87,13 @@ fn gather_share_files() -> Result<Vec<PathBuf>> {
         ret.push(entry_path.into());
     }
 
+    if ret.is_empty() {
+        return Err(anyhow!(
+            "Unable to find any resources in `{}`",
+            base_path.display()
+        ));
+    }
+
     Ok(ret)
 }
 
@@ -101,7 +104,7 @@ fn gather_share_files() -> Result<Vec<PathBuf>> {
 /// those files and their contents to be embedded into the subplot binary.
 fn write_out_resource_file<'a>(paths: impl Iterator<Item = &'a Path>) -> Result<()> {
     let mut out_path: PathBuf = std::env::var("OUT_DIR")?.into();
-    out_path.push("embedded_files.inc");
+    out_path.push("embedded_files.rs");
     let mut fh = std::fs::File::create(out_path)?;
     writeln!(fh, "&[")?;
     for entry in paths {
@@ -138,13 +141,7 @@ fn adopt_env_var(var: &str, def: &str) {
 
 fn main() {
     println!("cargo:rerun-if-env-changed=DEB_BUILD_OPTIONS");
-    let paths = if std::env::var("DEB_BUILD_OPTIONS").is_err() {
-        println!("cargo:rustc-env=FALLBACK_PATH=");
-        gather_share_files().expect("Unable to scan the share tree")
-    } else {
-        println!("cargo:rustc-env=FALLBACK_PATH=/usr/share/subplot");
-        vec![]
-    };
+    let paths = gather_share_files().expect("Unable to scan the share tree");
     write_out_resource_file(paths.iter().map(PathBuf::as_path))
         .expect("Unable to write the resource file out");
     adopt_env_var("DOT_PATH", "dot");
