@@ -93,10 +93,10 @@ def runcmd_helper_srcdir_path(ctx):
 
 
 # Step: This creates a helper script.
-def runcmd_helper_script(ctx, filename=None):
+def runcmd_helper_script(ctx, script=None):
     get_file = globals()["get_file"]
-    with open(filename, "wb") as f:
-        f.write(get_file(filename))
+    with open(script, "wb") as f:
+        f.write(get_file(script))
 
 
 #
