@@ -9,11 +9,15 @@ use serde::{Deserialize, Serialize};
 use serde_aux::prelude::*;
 
 use std::collections::HashMap;
+use std::fmt::Debug;
+use std::ops::Deref;
 use std::path::Path;
 use std::str::FromStr;
+use std::sync::Arc;
 
 use lazy_static::lazy_static;
 use regex::{escape, Regex, RegexBuilder};
+use tracing::{event, instrument, Level};
 
 #[derive(Debug, PartialEq, Eq, Hash, Copy, Clone, Serialize, Deserialize)]
 #[serde(rename_all = "lowercase")]
@@ -88,6 +92,61 @@ impl CaptureType {
         }
     }
 }
+
+/// A link from a binding to its implementation in a given language.
+///
+/// Such a link comprises a function name to call for the step and
+/// an optional function name to call to clean the step up at the end.
+#[derive(Debug, Clone)]
+pub struct BindingImpl {
+    function: String,
+    cleanup: Option<String>,
+}
+
+impl BindingImpl {
+    /// Create a new binding implementation
+    ///
+    /// ```ignore
+    /// # use subplot::bindings::BindingImpl;
+    ///
+    /// let bimpl = BindingImpl::new("foo::bar::func", Some("foo::bar::func_cleanup"));
+    /// ```
+    pub fn new(function: &str, cleanup: Option<&str>) -> Self {
+        Self {
+            function: function.to_string(),
+            cleanup: cleanup.map(str::to_string),
+        }
+    }
+
+    /// Retrieve the function name in this binding
+    ///
+    /// ```ignore
+    /// # use subplot::bindings::BindingImpl;
+    /// let bimpl = BindingImpl::new("foo::bar::func", None);
+    ///
+    /// assert_eq!(bimpl.function(), "foo::bar::func");
+    /// ```
+    pub fn function(&self) -> &str {
+        &self.function
+    }
+
+    /// Retrieve the cleanup function name in this binding
+    ///
+    /// ```ignore
+    /// # use subplot::bindings::BindingImpl;
+    /// let bimpl = BindingImpl::new("foo::bar::func", None);
+    ///
+    /// assert_eq!(bimpl.cleanup(), None);
+    ///
+    /// let bimpl = BindingImpl::new("foo::bar::func", Some("foo::bar::func_cleanup"));
+    ///
+    /// assert_eq!(bimpl.cleanup(), Some("foo::bar::func_cleanup"));
+    /// ```
+    pub fn cleanup(&self) -> Option<&str> {
+        self.cleanup.as_deref()
+    }
+}
+
 /// A binding of a scenario step to its implementation.
 ///
 /// Contains the pattern used to match against scenario steps,
@@ -98,8 +157,7 @@ pub struct Binding {
     kind: StepKind,
     pattern: String,
     regex: Regex,
-    function: String,
-    cleanup: Option<String>,
+    impls: HashMap<String, Arc<BindingImpl>>,
     types: HashMap<String, CaptureType>,
 }
 
@@ -108,8 +166,6 @@ impl Binding {
     pub fn new(
         kind: StepKind,
         pattern: &str,
-        function: &str,
-        cleanup: Option<&str>,
         case_sensitive: bool,
         mut types: HashMap<String, CaptureType>,
     ) -> Result<Binding> {
@@ -127,12 +183,19 @@ impl Binding {
             kind,
             pattern: pattern.to_owned(),
             regex,
-            function: function.to_string(),
-            cleanup: cleanup.map(String::from),
+            impls: HashMap::new(),
             types,
         })
     }
 
+    /// Insert an impl into this binding
+    pub fn add_impl(&mut self, template: &str, function: &str, cleanup: Option<&str>) {
+        self.impls.insert(
+            template.to_string(),
+            Arc::new(BindingImpl::new(function, cleanup)),
+        );
+    }
+
     /// Return the kind of step the binding is for.
     pub fn kind(&self) -> StepKind {
         self.kind
@@ -143,14 +206,9 @@ impl Binding {
         &self.pattern
     }
 
-    /// Return name of function that implements step.
-    pub fn function(&self) -> &str {
-        &self.function
-    }
-
-    /// Return name of function that implements cleanup.
-    pub fn cleanup(&self) -> Option<&str> {
-        self.cleanup.as_deref()
+    /// Retrieve a particular implementation by name
+    pub fn step_impl(&self, template: &str) -> Option<Arc<BindingImpl>> {
+        self.impls.get(template).cloned()
     }
 
     /// Return the compiled regular expression for the pattern of the
@@ -161,8 +219,13 @@ impl Binding {
         &self.regex
     }
 
+    /// Return the type bindings for this binding.
+    pub fn types(&self) -> impl Iterator<Item = (&str, CaptureType)> {
+        self.types.iter().map(|(s, c)| (s.as_str(), *c))
+    }
+
     /// Try to match defined binding against a parsed scenario step.
-    pub fn match_with_step(&self, step: &ScenarioStep) -> Option<MatchedStep> {
+    pub fn match_with_step(&self, template: &str, step: &ScenarioStep) -> Option<MatchedStep> {
         if self.kind() != step.kind() {
             return None;
         }
@@ -171,7 +234,7 @@ impl Binding {
         let caps = self.regex.captures(step_text)?;
 
         // If there is only one capture, it's the whole string.
-        let mut m = MatchedStep::new(self, &self.types);
+        let mut m = MatchedStep::new(self, template);
         if caps.len() == 1 {
             m.append_part(PartialStep::uncaptured(step_text));
             return Some(m);
@@ -246,82 +309,27 @@ mod test_binding {
     use std::collections::HashMap;
 
     #[test]
-    fn creates_new_without_cleanup() {
-        let b = Binding::new(
-            StepKind::Given,
-            "I am Tomjon",
-            "set_name",
-            None,
-            false,
-            HashMap::new(),
-        )
-        .unwrap();
+    fn creates_new() {
+        let b = Binding::new(StepKind::Given, "I am Tomjon", false, HashMap::new()).unwrap();
         assert_eq!(b.kind(), StepKind::Given);
         assert!(b.regex().is_match("I am Tomjon"));
         assert!(!b.regex().is_match("I am Tomjon of Lancre"));
         assert!(!b.regex().is_match("Hello, I am Tomjon"));
-        assert_eq!(b.function(), "set_name");
-        assert_eq!(b.cleanup(), None);
-    }
-
-    #[test]
-    fn creates_new_with_cleanup() {
-        let b = Binding::new(
-            StepKind::Given,
-            "I am Tomjon",
-            "set_name",
-            Some("unset_name"),
-            false,
-            HashMap::new(),
-        )
-        .unwrap();
-        assert_eq!(b.kind(), StepKind::Given);
-        assert!(b.regex().is_match("I am Tomjon"));
-        assert!(!b.regex().is_match("I am Tomjon of Lancre"));
-        assert!(!b.regex().is_match("Hello, I am Tomjon"));
-        assert_eq!(b.function(), "set_name");
-        assert_eq!(b.cleanup(), Some("unset_name"));
     }
 
     #[test]
     fn equal() {
-        let a = Binding::new(
-            StepKind::Given,
-            "I am Tomjon",
-            "set_name",
-            Some("unset"),
-            false,
-            HashMap::new(),
-        )
-        .unwrap();
-        let b = Binding::new(
-            StepKind::Given,
-            "I am Tomjon",
-            "set_name",
-            Some("unset"),
-            false,
-            HashMap::new(),
-        )
-        .unwrap();
+        let a = Binding::new(StepKind::Given, "I am Tomjon", false, HashMap::new()).unwrap();
+        let b = Binding::new(StepKind::Given, "I am Tomjon", false, HashMap::new()).unwrap();
         assert_eq!(a, b);
     }
 
     #[test]
     fn not_equal() {
-        let a = Binding::new(
-            StepKind::Given,
-            "I am Tomjon",
-            "set_name",
-            None,
-            false,
-            HashMap::new(),
-        )
-        .unwrap();
+        let a = Binding::new(StepKind::Given, "I am Tomjon", false, HashMap::new()).unwrap();
         let b = Binding::new(
             StepKind::Given,
             "I am Tomjon of Lancre",
-            "set_name",
-            None,
             false,
             HashMap::new(),
         )
@@ -332,30 +340,22 @@ mod test_binding {
     #[test]
     fn does_not_match_with_wrong_kind() {
         let step = ScenarioStep::new(StepKind::Given, "given", "yo");
-        let b = Binding::new(StepKind::When, "yo", "do_yo", None, false, HashMap::new()).unwrap();
-        assert!(b.match_with_step(&step).is_none());
+        let b = Binding::new(StepKind::When, "yo", false, HashMap::new()).unwrap();
+        assert!(b.match_with_step("", &step).is_none());
     }
 
     #[test]
     fn does_not_match_with_wrong_text() {
         let step = ScenarioStep::new(StepKind::Given, "given", "foo");
-        let b = Binding::new(StepKind::Given, "bar", "yo", None, false, HashMap::new()).unwrap();
-        assert!(b.match_with_step(&step).is_none());
+        let b = Binding::new(StepKind::Given, "bar", false, HashMap::new()).unwrap();
+        assert!(b.match_with_step("", &step).is_none());
     }
 
     #[test]
     fn match_with_fixed_pattern() {
         let step = ScenarioStep::new(StepKind::Given, "given", "foo");
-        let b = Binding::new(
-            StepKind::Given,
-            "foo",
-            "do_foo",
-            None,
-            false,
-            HashMap::new(),
-        )
-        .unwrap();
-        let m = b.match_with_step(&step).unwrap();
+        let b = Binding::new(StepKind::Given, "foo", false, HashMap::new()).unwrap();
+        let m = b.match_with_step("", &step).unwrap();
         assert_eq!(m.kind(), StepKind::Given);
         let mut parts = m.parts();
         let p = parts.next().unwrap();
@@ -369,13 +369,11 @@ mod test_binding {
         let b = Binding::new(
             StepKind::Given,
             r"I am (?P<who>\S+), I am",
-            "set_name",
-            None,
             false,
             HashMap::new(),
         )
         .unwrap();
-        let m = b.match_with_step(&step).unwrap();
+        let m = b.match_with_step("", &step).unwrap();
         assert_eq!(m.kind(), StepKind::Given);
         let mut parts = m.parts();
         assert_eq!(parts.next().unwrap(), &PartialStep::uncaptured("I am "));
@@ -387,26 +385,10 @@ mod test_binding {
     #[test]
     fn case_sensitive_mismatch() {
         let step = ScenarioStep::new(StepKind::Given, "given", "I am Tomjon");
-        let b = Binding::new(
-            StepKind::Given,
-            r"i am tomjon",
-            "set_name",
-            None,
-            false,
-            HashMap::new(),
-        )
-        .unwrap();
-        assert!(b.match_with_step(&step).is_some());
-        let b = Binding::new(
-            StepKind::Given,
-            r"i am tomjon",
-            "set_name",
-            None,
-            true,
-            HashMap::new(),
-        )
-        .unwrap();
-        assert!(b.match_with_step(&step).is_none());
+        let b = Binding::new(StepKind::Given, r"i am tomjon", false, HashMap::new()).unwrap();
+        assert!(b.match_with_step("", &step).is_some());
+        let b = Binding::new(StepKind::Given, r"i am tomjon", true, HashMap::new()).unwrap();
+        assert!(b.match_with_step("", &step).is_none());
     }
 }
 
@@ -422,13 +404,34 @@ impl Default for Bindings {
     }
 }
 
+#[derive(Debug, Deserialize)]
+struct ParsedImpl {
+    function: String,
+    cleanup: Option<String>,
+}
+
+#[derive(Debug, Deserialize)]
+#[serde(transparent)]
+struct ParsedImplWrapper {
+    #[serde(deserialize_with = "deserialize_struct_case_insensitive")]
+    pimpl: ParsedImpl,
+}
+
+impl Deref for ParsedImplWrapper {
+    type Target = ParsedImpl;
+
+    fn deref(&self) -> &Self::Target {
+        &self.pimpl
+    }
+}
+
 #[derive(Debug, Deserialize)]
 struct ParsedBinding {
     given: Option<String>,
     when: Option<String>,
     then: Option<String>,
-    function: String,
-    cleanup: Option<String>,
+    #[serde(default, rename = "impl")]
+    impls: HashMap<String, ParsedImplWrapper>,
     regex: Option<bool>,
     #[serde(default)]
     case_sensitive: bool,
@@ -465,6 +468,7 @@ impl Bindings {
     }
 
     /// Add bindings from a YAML string
+    #[instrument(level = "trace", skip(self, yaml))]
     pub fn add_from_yaml(&mut self, yaml: &str) -> Result<()> {
         let bindings: Vec<ParsedBindingWrapper> = serde_yaml::from_str(yaml)?;
         for wrapper in bindings {
@@ -480,11 +484,11 @@ impl Bindings {
 
     /// Find the binding matching a given scenario step, if there is
     /// exactly one.
-    pub fn find(&self, step: &ScenarioStep) -> Result<MatchedStep> {
+    pub fn find(&self, template: &str, step: &ScenarioStep) -> Result<MatchedStep> {
         let mut matches: Vec<MatchedStep> = self
             .bindings()
             .iter()
-            .filter_map(|b| b.match_with_step(step))
+            .filter_map(|b| b.match_with_step(template, step))
             .collect();
         if matches.len() > 1 {
             // Too many matching bindings.
@@ -502,13 +506,17 @@ impl Bindings {
     }
 
     /// Add bindings from a file.
-    pub fn add_from_file<P>(&mut self, filename: P) -> Result<()>
+    #[instrument(level = "trace", skip(self))]
+    pub fn add_from_file<P>(&mut self, filename: P, template: Option<&str>) -> Result<()>
     where
-        P: AsRef<Path>,
+        P: AsRef<Path> + Debug,
     {
-        let yaml = resource::read_as_string(filename.as_ref())
+        let yaml = resource::read_as_string(filename.as_ref(), template)
             .map_err(|e| SubplotError::BindingsFileNotFound(filename.as_ref().into(), e))?;
-        self.add_from_yaml(&yaml)?;
+        event!(Level::TRACE, "Loaded file content");
+        self.add_from_yaml(&yaml).map_err(|e| {
+            SubplotError::BindingFileParseError(filename.as_ref().to_owned(), Box::new(e))
+        })?;
         Ok(())
     }
 
@@ -522,6 +530,7 @@ impl Bindings {
     }
 }
 
+#[instrument(level = "trace", skip(parsed))]
 fn from_hashmap(parsed: &ParsedBinding) -> Result<Binding> {
     let given: i32 = parsed.given.is_some().into();
     let when: i32 = parsed.when.is_some().into();
@@ -557,14 +566,21 @@ fn from_hashmap(parsed: &ParsedBinding) -> Result<Binding> {
         regex_from_simple_pattern(pattern, parsed.regex.is_some(), &mut types)?
     };
 
-    Binding::new(
-        kind,
-        &pattern,
-        &parsed.function,
-        parsed.cleanup.as_deref(),
-        parsed.case_sensitive,
-        types,
-    )
+    event!(
+        Level::TRACE,
+        ?kind,
+        ?pattern,
+        ?types,
+        "Successfully acquired binding"
+    );
+
+    let mut ret = Binding::new(kind, &pattern, parsed.case_sensitive, types)?;
+    event!(Level::TRACE, "Binding parsed OK");
+    for (template, pimpl) in &parsed.impls {
+        ret.add_impl(template, &pimpl.function, pimpl.cleanup.as_deref());
+    }
+
+    Ok(ret)
 }
 
 #[cfg(test)]
@@ -589,8 +605,6 @@ mod test_bindings {
         let binding = Binding::new(
             StepKind::Given,
             r"I am (?P<name>\S+)",
-            "set_name",
-            None,
             false,
             HashMap::new(),
         )
@@ -604,21 +618,31 @@ mod test_bindings {
     fn adds_from_yaml() {
         let yaml = "
 - GIVEN: I am Tomjon
-  function: set_name
+  impl:
+    python:
+      function: set_name
 - when: I declare myself king
-  Function: declare_king
+  impl:
+    python:
+      Function: declare_king
 - tHEn: there is applause
-  function: check_for_applause
+  impl:
+    python:
+      function: check_for_applause
 - given: you are alice
-  function: other_name
+  impl:
+    python:
+      function: other_name
   case_sensitive: true
 - then: the total is {total}
-  function: check_total
+  impl:
+    python:
+      function: check_total
   types:
     total: word
 ";
         let mut bindings = Bindings::new();
-        bindings.add_from_yaml(&yaml).unwrap();
+        bindings.add_from_yaml(yaml).unwrap();
         println!("test: {:?}", bindings);
         assert!(bindings.has(StepKind::Given, "I am Tomjon"));
         assert!(bindings.has(StepKind::When, "I declare myself king"));
@@ -634,9 +658,11 @@ mod test_bindings {
         let yaml = "
 - Given: I am Tomjon
   wheN: I am indeed Tomjon
-  FUNCTION: set_name
+  impl:
+    python:
+      FUNCTION: set_name
 ";
-        match Bindings::new().add_from_yaml(&yaml) {
+        match Bindings::new().add_from_yaml(yaml) {
             Ok(_) => unreachable!(),
             Err(SubplotError::BindingHasManyKeywords(_)) => (),
             Err(e) => panic!("Incorrect error: {}", e),
@@ -647,11 +673,13 @@ mod test_bindings {
     fn typemap_must_match_pattern() {
         let yaml = "
 - then: you are {age:word} years old
-  function: check_age
+  impl:
+    python:
+      function: check_age
   types:
     age: number
 ";
-        match Bindings::new().add_from_yaml(&yaml) {
+        match Bindings::new().add_from_yaml(yaml) {
             Ok(_) => unreachable!(),
             Err(SubplotError::SimplePatternKindMismatch(_)) => (),
             Err(e) => panic!("Incorrect error: {}", e),
@@ -661,19 +689,11 @@ mod test_bindings {
     #[test]
     fn does_not_find_match_for_unmatching_kind() {
         let step = ScenarioStep::new(StepKind::Given, "given", "I am Tomjon");
-        let binding = Binding::new(
-            StepKind::When,
-            r"I am Tomjon",
-            "set_foo",
-            None,
-            false,
-            HashMap::new(),
-        )
-        .unwrap();
+        let binding = Binding::new(StepKind::When, r"I am Tomjon", false, HashMap::new()).unwrap();
         let mut bindings = Bindings::new();
         bindings.add(binding);
         assert!(matches!(
-            bindings.find(&step),
+            bindings.find("", &step),
             Err(SubplotError::BindingUnknown(_))
         ));
     }
@@ -684,8 +704,6 @@ mod test_bindings {
         let binding = Binding::new(
             StepKind::Given,
             r"I am Tomjon of Lancre",
-            "set_foo",
-            None,
             false,
             HashMap::new(),
         )
@@ -693,7 +711,7 @@ mod test_bindings {
         let mut bindings = Bindings::new();
         bindings.add(binding);
         assert!(matches!(
-            bindings.find(&step),
+            bindings.find("", &step),
             Err(SubplotError::BindingUnknown(_))
         ));
     }
@@ -702,31 +720,19 @@ mod test_bindings {
     fn two_matching_bindings() {
         let step = ScenarioStep::new(StepKind::Given, "given", "I am Tomjon");
         let mut bindings = Bindings::default();
-        bindings.add(
-            Binding::new(
-                StepKind::Given,
-                r"I am Tomjon",
-                "set_foo",
-                None,
-                false,
-                HashMap::new(),
-            )
-            .unwrap(),
-        );
+        bindings.add(Binding::new(StepKind::Given, r"I am Tomjon", false, HashMap::new()).unwrap());
         bindings.add(
             Binding::new(
                 StepKind::Given,
                 &super::regex_from_simple_pattern(r"I am {name}", false, &mut HashMap::new())
                     .unwrap(),
-                "set_foo",
-                None,
                 false,
                 HashMap::new(),
             )
             .unwrap(),
         );
         assert!(matches!(
-            bindings.find(&step),
+            bindings.find("", &step),
             Err(SubplotError::BindingNotUnique(_, _))
         ));
     }
@@ -734,18 +740,10 @@ mod test_bindings {
     #[test]
     fn finds_match_for_fixed_string_pattern() {
         let step = ScenarioStep::new(StepKind::Given, "given", "I am Tomjon");
-        let binding = Binding::new(
-            StepKind::Given,
-            r"I am Tomjon",
-            "set_name",
-            None,
-            false,
-            HashMap::new(),
-        )
-        .unwrap();
+        let binding = Binding::new(StepKind::Given, r"I am Tomjon", false, HashMap::new()).unwrap();
         let mut bindings = Bindings::new();
         bindings.add(binding);
-        let m = bindings.find(&step).unwrap();
+        let m = bindings.find("", &step).unwrap();
         assert_eq!(m.kind(), StepKind::Given);
         let mut parts = m.parts();
         let p = parts.next().unwrap();
@@ -762,15 +760,13 @@ mod test_bindings {
         let binding = Binding::new(
             StepKind::Given,
             r"I am (?P<name>\S+)",
-            "set_name",
-            None,
             false,
             HashMap::new(),
         )
         .unwrap();
         let mut bindings = Bindings::new();
         bindings.add(binding);
-        let m = bindings.find(&step).unwrap();
+        let m = bindings.find("", &step).unwrap();
         assert_eq!(m.kind(), StepKind::Given);
         let mut parts = m.parts();
         let p = parts.next().unwrap();
