@@ -36,6 +36,10 @@ pub enum SubplotError {
     #[error("simple pattern contains regex metacharacters: {0}")]
     SimplePatternHasMetaCharacters(String),
 
+    /// Error while parsing a bindings file
+    #[error("binding file failed to parse: {0}")]
+    BindingFileParseError(PathBuf, #[source] Box<SubplotError>),
+
     /// Scenario step does not match a known binding
     ///
     /// This may be due to the binding missing entirely, or that the
@@ -161,7 +165,7 @@ pub enum SubplotError {
     ///
     /// Names of embedded files must be unique in the input document,
     /// but Subplot found at least one with the same name as another.
-    #[error("Duplicate embedded file name: {0}")]
+    #[error("Duplicate embedded file name: {0:?}")]
     DuplicateEmbeddedFilename(String),
 
     /// Embedded file has more than one `add-newline` attribute
@@ -229,6 +233,10 @@ pub enum SubplotError {
     #[error("failure rendering pikchr diagram: {0}")]
     PikchrRenderError(String),
 
+    /// When attempting to codegen, no scenarios matched the desired template language
+    #[error("no scenarios were found matching the `{0}` template")]
+    NoScenariosMatched(String),
+
     /// I/O error
     ///
     /// Subplot did some I/O, and it failed. This is a generic wrapper
@@ -283,6 +291,14 @@ pub enum SubplotError {
         #[from]
         source: serde_yaml::Error,
     },
+
+    /// Abstract syntax tree error.
+    #[error(transparent)]
+    Ast(#[from] crate::ast::Error),
+
+    /// UTF8 conversion error.
+    #[error(transparent)]
+    Utf8Error(#[from] std::str::Utf8Error),
 }
 
 impl SubplotError {
