@@ -3,6 +3,7 @@ use crate::DataFile;
 use crate::DataFiles;
 
 use pandoc_ast::{Block, MutVisitor};
+use tracing::{event, Level};
 
 impl MutVisitor for DataFiles {
     fn visit_vec_block(&mut self, vec_block: &mut Vec<Block>) {
@@ -11,7 +12,8 @@ impl MutVisitor for DataFiles {
             match block {
                 Block::CodeBlock(attr, contents) => {
                     if is_class(attr, "file") {
-                        let add_newline = match panhelper::find_attr_kv(&attr, "add-newline").next()
+                        event!(Level::TRACE, ?attr, "visiting code block");
+                        let add_newline = match panhelper::find_attr_kv(attr, "add-newline").next()
                         {
                             None | Some("auto") => !contents.ends_with('\n'),
                             Some("yes") => true,
